__version__ = "2023.6.1"

# flake8: noqa
from .cov_estimator import CovarianceEstimator, RollingCovarianceEstimator
from .factor_risk_model import FactorRiskModel
from .rolling_factor_risk_model import RollingFactorRiskModel



from datetime import datetime
from typing import Dict, Iterable, Optional, Tuple

from pandas import DataFrame, Timestamp

from .config import Config
from .risk_model import RiskModel


class RollingRiskModelConfig(Config):
    """
    Rollingrisk model configuration.

    Parameters
    ----------
    window: Optional[int]
        Number of rolling windows to use from the returns.
        Must be provided in fitting the model.

    show_progress: Optional[bool]
        Indicate to show progress bar in running.
    """

    window: int
    show_progress: bool = False


class RollingRiskModel:
    """
    Rolling risk model.

    A rolling risk model simply holds a dictionary of risk models
    in a given date / time range. Each key and value pair is a
    date / time and risk model object respectively.
    """

    ConfigClass = RollingRiskModelConfig

    def __init__(
        self,
        model: Optional[RiskModel] = None,
        window: Optional[int] = None,
        show_progress: Optional[bool] = False,
        values: Optional[Dict[datetime, RiskModel]] = None,
    ):
        """
        Constructor.

        Parameters
        ----------
        model: Optional[RiskModel]
            Risk model object to fit in rolling basis.

        window: Optional[int]
            Number of rolling windows to use from the returns.
            Must be provided in fitting the model.

        show_progress: Optional[bool]
            Indicate to show progress bar in running.

        values: Optional[Dict[datetime, RiskModel]]
            Rolling risk models values.
        """
        self._model = model
        self._values = values
        self._config = self.ConfigClass(window=window, show_progress=show_progress)

    @property
    def config(self):
        """
        Return the configuration object.
        """
        return self._config

    def get(self, name, **kwargs) -> RiskModel:
        """
        Return a factor risk model from the given name / key.
        """
        try:
            name = Timestamp(name)
        finally:
            return self._values.get(name, **kwargs)

    def keys(self) -> Iterable[datetime]:
        """
        Return a list of keys.
        """
        return self._values.keys()

    def values(self) -> Iterable[RiskModel]:
        """
        Return a list of values.
        """
        return self._values.values()

    def items(self) -> Iterable[Tuple[datetime, RiskModel]]:
        """
        Return a list of tuples with keys and values.
        """
        return self._values.items()

    def fit(
        self,
        X: DataFrame,
        validity: Optional[DataFrame] = None,
        weights: Optional[DataFrame] = None,
    ) -> object:
        """
        Fit the model.

        Parameters
        ----------
        X: DataFrame
            The instrument returns of which its index and columns
            are the date / time and return values.

        validity: DataFrame
            The instrument validity on the date.

        weights: DataFrame
            The weights of the instruments, same dimension as the
            instrument returns.

        Returns
        -------
        object
            The object itself.
        """
        values = {}

        T = X.shape[0]
        iterator = range(0, T)
        if self._config.show_progress:
            from tqdm import tqdm

            iterator = tqdm(iterator, leave=False)

        try:
            for index in iterator:
                start_index = index
                end_index = index + self._config.window + 1
                if end_index > T:
                    break

                if isinstance(X, DataFrame):
                    X_input = X.iloc[start_index:end_index, :]
                    index_name = X.index[end_index - 1]
                else:
                    raise TypeError(f"Invalid type of X {X.__class__.__name__}")

                if weights is None:
                    weights_input = None
                elif isinstance(weights, DataFrame):
                    weights_input = weights.loc[index_name]
                else:
                    raise TypeError(
                        f"Invalid type of weights {weights.__class__.__name__}"
                    )

                if validity is not None:
                    validity_input = validity.loc[index_name]
                    X_input = X_input.loc[:, validity_input]

                if X_input.shape[1] == 0:
                    continue

                params = {}
                if weights_input is not None:
                    params["weights"] = weights_input

                values[index_name] = self._model.fit(
                    X=X_input.fillna(0.0), **params
                ).copy()
        except Exception as exc:
            raise RuntimeError(
                f"Failed to fit at the index {index} due to error: {exc}"
            )

        self._values = values
        return self

    def asdict(self):
        """
        Returns a dict representation of the object.
        """
        return self._config.dict()



import json
from datetime import datetime
from os import makedirs
from os.path import basename, dirname
from os.path import join as fsjoin
from typing import Any, Dict, Optional

import pandas as pd
from pandas import DataFrame

from ..factor_risk_model import FactorRiskModel
from ..rolling_factor_risk_model import RollingFactorRiskModel


def generate_factor_risk_model(
    model: str, data: DataFrame, **kwargs
) -> FactorRiskModel:
    """
    Generate factor risk model

    Parameters
    ----------
    model : str
      Model name supported in statistics module. Supported
      value is `pca`.

    data: DataFrame
      Dataframe of returns of valid instruments, in a dimension
      of (T, N) where N is the number of instruments and T is the
      of timeframes.

    Returns
    -------
    FactorRiskModel
      A fitted factor risk model.
    """
    model = model.lower().replace("-", "_")
    if model == "pca":
        from ..statistical.pca import PCA

        model = PCA(**kwargs)
    elif model == "apca":
        from ..statistical.apca import APCA

        model = APCA(**kwargs)
    else:
        raise ValueError(f"Model name {model} is not supported")

    return model.fit(X=data)


def generate_rolling_factor_risk_model(
    model: str,
    data: DataFrame,
    model_parameters: Dict[str, Any],
    weights: Optional[DataFrame] = None,
    **kwargs,
) -> RollingFactorRiskModel:
    model = model.lower().replace("-", "_")
    if model == "pca":
        from ..statistical.pca import PCA

        model = PCA(**model_parameters)
    elif model == "apca":
        from ..statistical.apca import APCA

        model = APCA(**model_parameters)
    else:
        raise ValueError(f"Model name {model} is not supported")
    rolling_model = RollingFactorRiskModel(model=model, **kwargs)

    params = {}
    if weights is not None:
        params["weights"] = weights

    return rolling_model.fit(X=data, **params)


def dump_factor_risk_model(
    risk_model: FactorRiskModel,
    metadata_file: str,
    format: str,
    parameters: Optional[Dict] = None,
):
    """
    Dump factor risk model.
    """
    parameters = parameters or {}
    dumper = f"to_{format}"
    output_directory = dirname(metadata_file)

    def _dump(name, data, output_directory):
        if isinstance(data, pd.DataFrame):
            makedirs(output_directory, exist_ok=True)
            getattr(data, dumper)(
                fsjoin(output_directory, f"{name}.{format}"), **parameters
            )
        elif isinstance(data, dict):
            for key, value in data.items():
                _dump(
                    name=key,
                    data=value,
                    output_directory=fsjoin(output_directory, name),
                )
        else:
            raise TypeError(f"Unrecognised type {data.__class__.__class__} to export")

    _dump(
        name="factor-exposures",
        data=risk_model.factor_exposures,
        output_directory=output_directory,
    )

    _dump(
        name="factor-returns",
        data=risk_model.factor_returns,
        output_directory=output_directory,
    )

    _dump(
        name="residual-returns",
        data=risk_model.residual_returns,
        output_directory=output_directory,
    )

    with open(metadata_file, mode="w") as f:
        f.write(json.dumps({"parameters": risk_model.asdict()}))


def dump_rolling_factor_risk_model(
    rolling_risk_model: RollingFactorRiskModel,
    metadata_file: str,
    format: str,
    parameters: Optional[Dict] = None,
    show_progress: Optional[bool] = True,
):
    keys = []
    iterator = rolling_risk_model.items()
    if show_progress:
        from tqdm import tqdm

        iterator = tqdm(iterator, leave=False)
    for key, model in iterator:
        if not isinstance(key, (pd.Timestamp, datetime)):
            raise TypeError(
                f"Key {key} type must be either datetime / Timestamp, "
                f"rather than {key.__class__.__name__}"
            )
        key_name = key.isoformat()
        keys.append(key_name)
        dump_factor_risk_model(
            risk_model=model,
            metadata_file=fsjoin(
                dirname(metadata_file),
                key_name,
                basename(metadata_file),
            ),
            format=format,
            parameters=parameters,
        )

    with open(metadata_file, mode="w") as f:
        f.write(
            json.dumps({"directories": keys, "parameters": rolling_risk_model.asdict()})
        )


def load_factor_risk_model(
    metadata_file: str,
    format: str,
    parameters: Optional[Dict] = None,
):
    parameters = parameters or {}
    loader = getattr(pd, f"read_{format}")
    output_directory = dirname(metadata_file)

    def _load(name):
        output_path = fsjoin(output_directory, f"{name}.{format}")
        return loader(output_path, **parameters)

    factor_exposures = _load("factor-exposures")
    factor_returns = _load("factor-returns")
    residual_returns = _load("residual-returns")
    return FactorRiskModel(
        factor_exposures=factor_exposures,
        factor_returns=factor_returns,
        residual_returns=residual_returns,
    )


def load_rolling_factor_risk_model(
    metadata_file: str,
    format: str,
    parameters: Optional[Dict] = None,
    show_progress: Optional[bool] = True,
):
    with open(metadata_file) as f:
        metadata = json.load(f)
    try:
        directories = metadata["directories"]
    except KeyError:
        raise RuntimeError(
            "Failed to retrieve the list of directories from "
            f"the metadata file {metadata_file}. Please ensure "
            "the directory was exported as rolling factor risk "
            "model format"
        )

    if show_progress:
        from tqdm import tqdm

        directories = tqdm(directories, leave=False)

    output_directory = dirname(metadata_file)
    metadata_file_name = basename(metadata_file)
    values = {}
    for directory in directories:
        risk_model_metadata_file = fsjoin(
            output_directory,
            directory,
            metadata_file_name,
        )
        risk_model = load_factor_risk_model(
            metadata_file=risk_model_metadata_file, format=format, parameters=parameters
        )
        values[pd.Timestamp(directory)] = risk_model

    model_parameters = metadata.get("parameters", {})
    return RollingFactorRiskModel(values=values, **model_parameters)


def where_validity(
    validity: pd.DataFrame,
    data: pd.DataFrame,
    fillna: Any = None,
    ffill: Optional[bool] = False,
) -> pd.DataFrame:
    """
    Return the data for the given universe.

    Parameters
    ----------
    validity : pd.DataFrame
      Validity of the universe of which the index and columns are date / time
      and instrument names respectively.
    data: pd.DataFrame
      Data of which the index and columns are date / time and instrument names
      respectively.
    fillna: Any
      Handle nan values which includes data outside of the universe.
    ffill: Optional[bool]
      Indicates to forward fill the data. Default is `False`.

    Returns
    -------
    pd.DataFrame
      Dataframe containing the data for the given universe.
    """
    data = data.reindex_like(validity).where(validity)
    if ffill:
        data = data.ffill()
    if fillna is not None:
        data = data.fillna(fillna)
    return data



import io
from os import listdir
from os.path import isfile, join
from tempfile import TemporaryDirectory
from zipfile import ZipFile

import pandas as pd
import requests


def get_attr(data_all, field_name):
    """
    Get attribute for each value in the dictionary.

    :param data_all: Data in dict format of which its values
      are dataframes.
    """
    attr = {key: df[field_name] for key, df in data_all.items()}
    attr = pd.DataFrame(attr)
    attr = attr.ffill().groupby(attr.index.date).first()
    attr.index = pd.to_datetime(attr.index)
    attr = attr.groupby(attr.index.date).last()
    attr.index = pd.to_datetime(attr.index)
    return attr


def download_sample_data_estimation_universe():
    """
    Download sample data of estimation universe.

    :return: Dict of sample data with prices, returns, marketcap
      and volumes.
    """
    data = {}
    response = requests.get(
        "https://github.com/factorpricingmodel/factor-pricing-model-risk-model"
        "/raw/main/examples/notebook/crypto_estimation_universe_sample_data.zip",
        stream=True,
    )
    with ZipFile(io.BytesIO(response.content)) as zip_ref:
        with TemporaryDirectory() as temp_dir:
            zip_ref.extractall(temp_dir)
            files = [
                (join(temp_dir, f), f.replace(".csv", ""))
                for f in listdir(temp_dir)
                if isfile(join(temp_dir, f))
            ]

            for file_path, name in files:
                df = pd.read_csv(file_path)
                df = df.set_index("date")
                df.index = pd.to_datetime(df.index)
                data[name] = df

    prices = get_attr(data, "price")
    returns = prices.pct_change()
    volumes = get_attr(data, "total_volume")
    marketcap = get_attr(data, "market_cap")

    return {
        "prices": prices,
        "returns": returns,
        "volumes": volumes,
        "marketcap": marketcap,
    }


def download_sample_data_model_universe():
    """
    Download sample data of model universe.

    :return: Dict of sample data with prices, returns, volumes
      (in crypto) and volumes (in USD).
    """
    data = {}
    response = requests.get(
        "https://github.com/factorpricingmodel/factor-pricing-model-risk-model"
        "/raw/main/examples/notebook/crypto_model_universe_sample_data.zip",
        stream=True,
    )
    with ZipFile(io.BytesIO(response.content)) as zip_ref:
        with TemporaryDirectory() as temp_dir:
            zip_ref.extractall(temp_dir)
            price_data_dir = join(temp_dir, "Price-Data")
            files = [
                (join(price_data_dir, f), f.replace(".csv", ""))
                for f in listdir(price_data_dir)
                if isfile(join(price_data_dir, f))
            ]

            for file_path, name in files:
                df = pd.read_csv(file_path)
                df = df.set_index("Date")
                df.index = pd.to_datetime(df.index)
                data[name] = df

    prices = get_attr(data, "Adj Close")
    returns = prices.pct_change()
    volumes = get_attr(data, "Volume")
    volumes_usd = volumes * prices

    return {
        "prices": prices,
        "returns": returns,
        "volumes": volumes,
        "volumes_usd": volumes_usd,
    }



from numpy import array, ndarray, sum

from .wls import WLS


class EWLS(WLS):
    """
    Exponential weighted least squares regression.

    The weighting scheme is

    .. math::
        w_t = \\frac{2^{-\\frac{T-t){\\lambda}}}}{\\bar{w}}

    where lambda is the half-life parameter.
    """

    def __init__(self, half_life: float, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        half_life : float
            Half-life of the weighting decay.
        """
        super().__init__(**kwargs)
        self._half_life = half_life

    def fit(self, X: ndarray, y: ndarray):
        """
        Fit the coefficients by the data X and y.

        Parameters
        ----------
        X: ndarray
          Training data.
        y: ndarray
          Target values.
        weights: Optional[ndarray]
          Weightings in regressiond data. The dimension should be
          same as y.
        """
        T = y.shape[0]
        weights = array([2 ** (-(T - 1 - t) / self._half_life) for t in range(0, T)])
        weights /= sum(weights)
        return super().fit(X=X, y=y, weights=weights)



from typing import Any, Dict, Optional, Union

from numpy import nan, sqrt, sum
from pandas import DataFrame, Series

from ..rolling_factor_risk_model import RollingFactorRiskModel


def compute_standardized_returns(
    X: DataFrame,
    weights: DataFrame,
    rolling_risk_model: Optional[Union[RollingFactorRiskModel, Dict[Any, Any]]] = None,
    forecast_vols: Optional[Series] = None,
    cov_halflife: Optional[float] = None,
) -> Series:
    """
    Compute the standardized returns given the rolling risk model.

    Standardized return is defined as

    .. math::
        b_t = \\frac{r_t}{\\sigma_t}


    Parameters
    ----------
    X: ndarray
        The instrument forecast returns.
    weights: ndarray
        Weights of the instruments.
    rolling_risk_model: Union[RollingFactorRiskModel, Dict[Any, Any]]
        A rolling risk model object or dictionary of covariances of
        which the keys and values are dates and covariances.
    forecast_vols: Series
        The forecast volatility.
    cov_halflife: Optional[float]
        Halflife in computing covariances.

    Returns
    -------
    Series
        A timeseries of standardized returns.
    """
    b_t = Series(nan, index=weights.index)
    instruments = weights.columns
    for index, index_weights in weights.iterrows():
        returns = sum(X.loc[index, instruments] * index_weights)
        if forecast_vols is not None:
            vol = forecast_vols.loc[index]
        else:
            risk_model = rolling_risk_model.get(index)
            if risk_model is None:
                continue
            elif isinstance(risk_model, DataFrame):
                cov = risk_model
            else:
                cov = risk_model.cov(halflife=cov_halflife)
            cov = cov.reindex(index=instruments, columns=instruments).fillna(0.0).values
            vol = sqrt((cov @ index_weights) @ index_weights)
        b_t[index] = returns / vol

    return b_t


def compute_bias_statistics(
    X: DataFrame,
    weights: DataFrame,
    window: int,
    rolling_risk_model: Optional[Union[RollingFactorRiskModel, Dict[Any, Any]]] = None,
    forecast_vols: Optional[Series] = None,
    min_periods: Optional[int] = None,
    cov_halflife: Optional[float] = None,
) -> Series:
    """
    Compute the bias statistics.

    Standardized return is defined as

     .. math::
        b_t = \\frac{r_t}{\\sigma_t}

    and the bias statistic is expressed as

     .. math::
        B_T(t) = \\sqrt{\\frac{1}{T} \\sigma_{\\tau}(b_{\\tau} - \\bar{b})^2}

    Parameters
    ----------
    X: ndarray
        The instrument forecast returns.
    weights: ndarray
        Weights of the instruments.
    forecast_vols: Optional[Series]
        The forecast volatility.
    rolling_risk_model: Union[RollingFactorRiskModel, Dict[Any, Any]]
        A rolling risk model object or dictionary of covariances of
        which the keys and values are dates and covariances.
    cov_halflife: Optional[float]
        Halflife in computing covariances.

    Returns
    -------
    Series
        A timeseries of bias statistic.
    """
    standardized_returns = compute_standardized_returns(
        X=X,
        weights=weights,
        rolling_risk_model=rolling_risk_model,
        forecast_vols=forecast_vols,
        cov_halflife=cov_halflife,
    )
    return standardized_returns.rolling(window, min_periods=min_periods).std()



from typing import Any, Dict, Optional, Union

from numpy import nan, ndarray, sqrt, sum
from pandas import DataFrame, Series
from scipy.stats import norm

from ..rolling_factor_risk_model import RollingFactorRiskModel


def compute_value_at_risk_threshold(
    weights: DataFrame,
    rolling_risk_model: Optional[Union[RollingFactorRiskModel, Dict[Any, Any]]] = None,
    forecast_vols: Optional[Series] = None,
    threshold: Optional[float] = 0.95,
    cov_halflife: Optional[float] = None,
) -> Series:
    """
    Compute the VaR threshold for the given weights.

    Parameters
    ----------
    weights : DataFrame
        The portfolio weights for each instrument. The input
        index and columns are the date / time and instruments
        respectively. The weights should be normalized, i.e.
        sum to one for each time frame.

    rolling_risk_model: Union[RollingFactorRiskModel, Dict[Any, Any]]
        A rolling risk model object or dictionary of covariances of
        which the keys and values are dates and covariances.

    forecast_vols: Optional[Series]
        The forecast volatility of instruments.

    threshold: float
        The threshold for the VaR. The value should be between 0
        and 1. Default is 95%.

    cov_halflife: Optional[float]
        Halflife in computing covariances.
    """
    if not (0.0 < threshold < 1.0):
        raise ValueError(f"Threshold {threshold} should be between 0 and 1")
    quantile = norm.ppf(threshold)
    value_at_risk = Series(nan, index=weights.index)
    instruments = weights.columns
    for index, index_weights in weights.iterrows():
        if rolling_risk_model is not None:
            risk_model = rolling_risk_model.get(index)
            if risk_model is None:
                continue
            elif isinstance(risk_model, DataFrame):
                cov = risk_model
            else:
                cov = risk_model.cov(halflife=cov_halflife)

            cov = cov.reindex(index=instruments, columns=instruments).fillna(0.0).values
            vol = sqrt((cov @ index_weights) @ index_weights)
        else:
            vol = forecast_vols[index]
        value_at_risk[index] = quantile * vol

    return value_at_risk


def compute_value_at_risk_breach_statistics(
    X: DataFrame,
    weights: DataFrame,
    rolling_risk_model: Optional[Union[RollingFactorRiskModel, Dict[Any, Any]]] = None,
    forecast_vols: Optional[Series] = None,
    threshold: Optional[float] = 0.95,
    cov_halflife: Optional[float] = None,
) -> Series:
    """
    Compute the VaR breach statistics.

    Parameters
    ----------
    X: DataFrame
        Instrument returns. The input index and columns are
        date / time and instruments respectively.

    weights : DataFrame
        The portfolio weights for each instrument. The input
        index and columns are the date / time and instruments
        respectively. The weights should be normalized, i.e.
        sum to one for each time frame.

    rolling_risk_model: Union[RollingFactorRiskModel, Dict[Any, Any]]
        A rolling risk model object or dictionary of covariances of
        which the keys and values are dates and covariances.

    forecast_vols: Optional[Series]
        The forecast volatility of instruments.

    threshold: float
        The threshold for the VaR. The value should be between 0
        and 1. Default is 95%.

    cov_halflife: Optional[float]
        Halflife in computing covariances.
    """
    value_at_risk_threshold = compute_value_at_risk_threshold(
        weights=weights,
        rolling_risk_model=rolling_risk_model,
        forecast_vols=forecast_vols,
        threshold=threshold,
        cov_halflife=cov_halflife,
    )
    portfolio_returns = sum(X * weights, axis=1)[value_at_risk_threshold.index]
    return portfolio_returns.le(-value_at_risk_threshold)


def compute_value_at_risk_rolling_breach_statistics(
    X: ndarray,
    weights: ndarray,
    window: int,
    rolling_risk_model: Optional[Union[RollingFactorRiskModel, Dict[Any, Any]]] = None,
    forecast_vols: Optional[Series] = None,
    threshold: Optional[float] = 0.95,
    min_periods: Optional[int] = None,
    cov_halflife: Optional[float] = None,
) -> Series:
    """
    Compute the VaR breach statistics.

    Parameters
    ----------
    X: DataFrame
        Instrument returns. The input index and columns are
        date / time and instruments respectively.

    weights : DataFrame
        The portfolio weights for each instrument. The input
        index and columns are the date / time and instruments
        respectively. The weights should be normalized, i.e.
        sum to one for each time frame.

    window: int
        The number of rolling time frames to compute the percentage
        of returns breaching the specified VaR.

    rolling_risk_model: Union[RollingFactorRiskModel, Dict[Any, Any]]
        A rolling risk model object or dictionary of covariances of
        which the keys and values are dates and covariances.

    forecast_vols: Optional[Series]
        The forecast volatility of instruments.

    threshold: Optional[float]
        The threshold for the VaR. The value should be between 0
        and 1. Default is 95%.

    cov_halflife: Optional[float]
        Halflife in computing covariances.
    """
    breach_statistics = compute_value_at_risk_breach_statistics(
        X=X,
        weights=weights,
        rolling_risk_model=rolling_risk_model,
        forecast_vols=forecast_vols,
        threshold=threshold,
        cov_halflife=cov_halflife,
    )
    return breach_statistics.rolling(window, min_periods=min_periods).sum() / window
